# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# Run from project folder with: python -m pca_b_stream.test.test

import random as rndm
import sys as syst
from math import prod as Product
from typing import Any, Tuple

import numpy as nmpy
import tqdm

import pca_b_stream as pcas


N_TESTS = 1000

byte_orders = ("=", "<", ">")
elm_types = pcas.main.VALID_NUMPY_TYPES
enumeration_orders = ("C", "F")

dim_range = (1, 3)
length_range = (10, 100)
n_value_range = (1, 4)


def Main() -> None:
    """"""
    if syst.argv.__len__() < 2:
        n_tests = N_TESTS
    else:
        try:
            n_tests = int(syst.argv[1])
        except ValueError:
            n_tests = N_TESTS

    print("A Few simple tests")

    _CheckEncoding(nmpy.zeros((10, 10), dtype=nmpy.bool_))
    _CheckEncoding(nmpy.ones((10, 10), dtype=nmpy.bool_))
    _CheckEncoding(nmpy.zeros((10, 10), dtype=nmpy.uint8))
    _CheckEncoding(nmpy.ones((10, 10), dtype=nmpy.uint8))
    _CheckEncoding(nmpy.zeros((10, 10), dtype=nmpy.int64))
    _CheckEncoding(nmpy.ones((10, 10), dtype=nmpy.int64))
    _CheckEncoding(nmpy.full((10, 10), 2, dtype=nmpy.int64))
    inp_array = nmpy.empty((10, 10), dtype=nmpy.uint8)
    for row in (0, 1, 8, 9):
        for col in (0, 1, 8, 9):
            inp_array.fill(0)
            inp_array[row, col] = 1
            _CheckEncoding(inp_array)

    print(f"Random tests: {n_tests}... [some might be invalidated]")

    n_valid_masks = 0

    for _ in tqdm.trange(n_tests, ncols=80, mininterval=2, colour="#00aa00"):
        elm_type = rndm.choice(elm_types)
        if nmpy.dtype(elm_type).byteorder == "|":
            full_dtype = "|" + elm_type
            dtype = elm_type
        else:
            full_dtype = rndm.choice(byte_orders) + elm_type
            dtype = full_dtype
        enumeration_order = rndm.choice(enumeration_orders)

        dimension = rndm.randint(*dim_range)
        shape = tuple(rndm.randint(*length_range) for _ in range(dimension))

        inp_array = nmpy.random.random(size=Product(shape))
        n_values = rndm.randint(*n_value_range)  # Other than zero
        if n_values > 1:
            inp_array = nmpy.digitize(
                inp_array, nmpy.linspace(0.0, 1.0, num=n_values + 2)
            )
        else:
            inp_array = inp_array > 0.5
        inp_array = inp_array.astype(dtype)
        inp_array = nmpy.reshape(inp_array, shape, order=enumeration_order)

        mask_is_valid, _ = pcas.PCAIsValid(inp_array)
        if not mask_is_valid:
            continue
        n_valid_masks += 1

        # nmpy.sctypeDict: see note on multiple-code types in main.py
        details = (
            int(nmpy.max(inp_array)),
            full_dtype[0],
            nmpy.sctypeDict[full_dtype[1]].__name__,
            enumeration_order,
            dimension,
            shape,
        )

        _CheckEncoding(inp_array, expected_details=details)

    print(f"Valid random tests: {n_valid_masks}/{n_tests}")


def _CheckEncoding(
    array: nmpy.ndarray, /, *, expected_details: Tuple[Any, ...] = None
) -> None:
    """"""
    encoding = pcas.PCA2BStream(array)
    decoding = pcas.BStream2PCA(encoding)
    assert nmpy.array_equal(array, decoding)

    if expected_details is not None:
        extracted_details = pcas.BStreamDetails(
            encoding, details="meTodl", should_print=False, should_return=True
        )
        _CheckThatDetailsMatch(extracted_details, expected_details)


def _CheckThatDetailsMatch(
    extracted_details: Tuple[Any, ...], expected_details: Tuple[Any, ...], /
) -> None:
    """
    Details must contain: "meTodl"
    """
    for idx, (elm_1, elm_2) in enumerate(zip(extracted_details, expected_details)):
        if idx == 1:
            if elm_2 == "=":
                mismatch = ((syst.byteorder == "big") and (elm_1 == "<")) or (
                    (syst.byteorder != "big") and (elm_1 == ">")
                )
            else:
                mismatch = elm_1 != elm_2
        elif idx == 3:
            if (extracted_details[4] > 1) and (expected_details[4] > 1):
                mismatch = elm_1 != elm_2
            else:
                mismatch = False
        else:
            mismatch = elm_1 != elm_2

        if mismatch:
            assert (
                extracted_details == expected_details
            ), f"Extracted={extracted_details}; Expected={expected_details}"


if __name__ == "__main__":
    #
    Main()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import leb128 as lenc

import numpy as nmpy

import base64 as bs64
import sys as syst
from typing import Any, Optional, Tuple
import zlib as cmpr


array_t = nmpy.ndarray


# See also: nmpy.sctypes and nmpy.sctypeDict
# /!\ Some types have several codes (e.g., "l" = "p" = numpy.int64; see mpy.sctypeDict)
VALID_NUMPY_TYPES = "?" + nmpy.typecodes["AllInteger"] + nmpy.typecodes["Float"]


def PCA2BStream(mask: array_t, /) -> bytes:
    """"""
    # --- Storage
    dtype = mask.dtype
    byte_order = dtype.byteorder
    if byte_order == "=":
        if syst.byteorder == "big":
            byte_order = ">"
        else:
            byte_order = "<"
    if mask.flags["C_CONTIGUOUS"]:
        enumeration_order = "C"
    else:
        enumeration_order = "F"
    storage = byte_order + dtype.char + enumeration_order

    # --- Geometry
    geometry = [mask.ndim, *mask.shape]

    # --- Contents
    raveled_mask = nmpy.ravel(mask, order="K")
    n_elements = raveled_mask.size
    values = range(max(1, int(nmpy.max(raveled_mask))) + 1)
    streams = []
    for v_idx, value in enumerate(values[1:]):
        if values.__len__() > 2:
            binary_mask = raveled_mask == value
            first_value_is_one = binary_mask[0]
        else:
            # Avoid additional array allocation when already binary
            binary_mask = raveled_mask
            first_value = binary_mask[0]
            first_value_is_one = (first_value == True) or (first_value > 0)

        # --- First value indicator
        if first_value_is_one:
            first_value = b"1"
        else:
            first_value = b"0"

        # --- Run lengths
        jumps = nmpy.diff(binary_mask)
        jump_idc = nmpy.nonzero(jumps)[0]
        if jump_idc.size > 1:
            run_lengths = (
                [jump_idc[0] + 1]
                + nmpy.diff(jump_idc).tolist()
                + [n_elements - jump_idc[-1] - 1]
            )
        elif jump_idc.size > 0:
            run_lengths = [jump_idc[0] + 1, n_elements - jump_idc[0] - 1]
        else:
            run_lengths = [n_elements]

        # --- Numeric encoding ([geometry + ]run lengths)
        numeric_encoding = []
        if v_idx > 0:
            numbers = run_lengths
        else:
            numbers = geometry + run_lengths
        for number in numbers:
            numeric_encoding.append(lenc.u.encode(number))
        numeric_encoding = b"".join(numeric_encoding)

        # --- Compression
        if v_idx > 0:
            uncompressed = first_value + numeric_encoding
        else:
            uncompressed = bytes(storage, "ascii") + first_value + numeric_encoding
        compressed = cmpr.compress(uncompressed, cmpr.Z_BEST_COMPRESSION)
        if compressed.__len__() > uncompressed.__len__():
            stream = b"0" + uncompressed
        else:
            stream = b"1" + compressed
        streams.append(bs64.b85encode(stream))

    return b"\n".join(streams)


def BStream2PCA(stream: bytes, /) -> array_t:
    """"""
    output = None

    dtype_as_str = None
    enumeration_order = None
    shape = None

    for s_idx, str_line in enumerate(stream.splitlines()):
        # --- Decompression
        decoded = bs64.b85decode(str_line)
        if chr(decoded[0]) == "0":
            decompressed = decoded[1:]
        else:
            decompressed = cmpr.decompress(decoded[1:])

        # --- Storage
        if s_idx > 0:
            first_value_idx = 0
        else:
            *dtype_as_str, enumeration_order = decompressed[:3]
            dtype_as_str = chr(dtype_as_str[0]) + chr(dtype_as_str[1])
            if dtype_as_str[0] == "|":
                dtype_as_str = dtype_as_str[1]
            enumeration_order = chr(enumeration_order)
            first_value_idx = 3

        # --- Numbers ([geometry + ]run lengths)
        numbers = []
        c_idx = first_value_idx + 1
        while c_idx < decompressed.__len__():
            last_idx = c_idx
            while decompressed[last_idx] & 0b10000000 > 0:
                last_idx += 1
            encoded_value = decompressed[c_idx : (last_idx + 1)]
            decoded_value = lenc.u.decode(encoded_value)
            numbers.append(decoded_value)
            c_idx = last_idx + 1

        # --- Geometry
        if s_idx > 0:
            run_lengths_idx = 0
        else:
            ndim = numbers[0]
            shape = numbers[1 : (ndim + 1)]
            run_lengths_idx = ndim + 1

        # --- Contents
        first_value = int(chr(decompressed[first_value_idx]))
        run_lengths = numbers[run_lengths_idx:]
        value_groups = (
            _lgt * (2 ** ((_idx + first_value) % 2) - 1,)
            for _idx, _lgt in enumerate(run_lengths)
        )
        values = tuple(_elm for _grp in value_groups for _elm in _grp)

        # --- Array creation
        if output is None:
            output = nmpy.array(values, dtype=dtype_as_str)
            output = nmpy.reshape(output, shape, order=enumeration_order)
        else:
            value = s_idx + 1
            mask = nmpy.array(values, dtype=dtype_as_str)
            mask = nmpy.reshape(mask, shape, order=enumeration_order)
            output[mask > 0] = value

    return output


def BStreamDetails(
    stream: bytes,
    /,
    *,
    details: str = "+",
    should_print: bool = True,
    should_return: bool = False,
) -> Optional[Any]:
    """
    Details: if "+", retrieve all details. Otherwise, pick among:
    m: maximum value in array (also number of sub-streams)
    c: compression indicator (string of zeros and ones, one per sub-stream)
    e: endianness (or byte order); See: https://numpy.org/doc/stable/reference/generated/numpy.dtype.byteorder.html
    t: dtype type code; See: https://numpy.org/doc/stable/reference/generated/numpy.dtype.char.html
    T: dtype name; Translated from "t" by Numpy.sctypeDict
    o: enumeration order; See: https://numpy.org/doc/stable/reference/generated/numpy.ndarray.flags.html, ?_CONTIGUOUS
    v: first value (0 for 0 or False, 1 for non-zero or True, one per sub-stream)
    d: array dimension
    l: array lengths per dimension
    """
    if not (should_print or should_return):
        return None

    if details == "+":
        output = {}
    else:
        output = {_key: None for _key in details}

    sub_streams = stream.splitlines()
    if ("m" in details) or (details == "+"):
        output["m"] = sub_streams.__len__()

    for s_idx, str_line in enumerate(sub_streams):
        # --- Decompression
        decoded = bs64.b85decode(str_line)
        compression_indicator = chr(decoded[0])
        if ("c" in details) or (details == "+"):
            if ("c" in output) and (output["c"] is not None):
                output["c"] = output["c"] + compression_indicator
            else:
                output["c"] = compression_indicator
        if compression_indicator == "0":
            decompressed = decoded[1:]
        else:
            decompressed = cmpr.decompress(decoded[1:])

        # --- Storage
        if s_idx > 0:
            first_value_idx = 0
        else:
            *dtype_as_str, enumeration_order = decompressed[:3]
            dtype_as_str = chr(dtype_as_str[0]) + chr(dtype_as_str[1])
            if ("e" in details) or (details == "+"):
                output["e"] = dtype_as_str[0]
            if ("t" in details) or (details == "+"):
                output["t"] = dtype_as_str[1]
            if ("T" in details) or (details == "+"):
                output["T"] = nmpy.sctypeDict[dtype_as_str[1]].__name__
            enumeration_order = chr(enumeration_order)
            if ("o" in details) or (details == "+"):
                output["o"] = enumeration_order
            first_value_idx = 3

        # --- Contents
        if ("v" in details) or (details == "+"):
            first_value = chr(decompressed[first_value_idx])
            if ("v" in output) and (output["v"] is not None):
                output["v"] = output["v"] + first_value
            else:
                output["v"] = first_value

        if (s_idx > 0) or (
            ("d" not in details) and ("l" not in details) and (details != "+")
        ):
            continue

        # --- Numbers ([geometry + ]run lengths)
        numbers = []
        c_idx = first_value_idx + 1
        expected_n_numbers = 1  # In fact, at least one
        while (c_idx < decompressed.__len__()) and (
            numbers.__len__() < expected_n_numbers
        ):
            last_idx = c_idx
            while decompressed[last_idx] & 0b10000000 > 0:
                last_idx += 1
            encoded_value = decompressed[c_idx : (last_idx + 1)]
            decoded_value = lenc.u.decode(encoded_value)
            if (numbers.__len__() == 0) and (("l" in details) or (details == "+")):
                expected_n_numbers = decoded_value + 1
            numbers.append(decoded_value)
            c_idx = last_idx + 1

        # --- Geometry
        if ("d" in details) or (details == "+"):
            output["d"] = numbers[0]
        if ("l" in details) or (details == "+"):
            output["l"] = tuple(numbers[1 : (numbers[0] + 1)])

    if should_print:
        for key, value in output.items():
            print(f"{key} = {value}")

    if should_return:
        if details == "+":
            return output
        else:
            output = tuple(output.values())
            if output.__len__() > 1:
                return output
            elif output.__len__() > 0:
                return output[0]
            else:
                return output  # Empty tuple

    return None


def PCAIsValid(mask: array_t, /) -> Tuple[bool, Optional[str]]:
    """"""
    issue = None

    if issue is None:
        if mask.dtype.char not in VALID_NUMPY_TYPES:
            issue = f"{mask.dtype.name}: Invalid type with code {mask.dtype.char}; Expected codes={VALID_NUMPY_TYPES}"

    if issue is None:
        unique_values = tuple(nmpy.unique(mask))
        if (unique_values != (False, True)) and (
            not set(unique_values).issubset(range(int(unique_values[-1]) + 1))
        ):
            issue = (
                "Mask is neither a boolean mask nor a numeric mask with integer values"
            )

    return issue is None, issue



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dominate.tags as html
from si_fi_o.generic import URLOfPath
from si_fi_o.session.session import session_t


_LEGEND = {
    "m": "Max. value in array (= number of sub-streams)",
    "c": "Compression indicator per sub-stream",
    "e": "Endianness (or byte order)",
    "t": "dtype code",
    "T": "dtype name",
    "o": "Enumeration order",
    "v": "First value per sub-stream (0: 0 or False, 1: non-zero or True)",
    "d": "Array dimension",
    "l": "Lengths per dimension",
}


def SessionInputsAsHTML(_: session_t | None, __: str, /) -> html.html_tag:
    """"""
    return html.div(_class="container")


def SessionOutputsAsHTML(session: session_t, /) -> html.html_tag | None:
    """
    Needs to be kept in sync with processing.py since it assigns the outputs
    """
    if (outputs := session.outputs) is None:
        return None

    stream, stream_size, details = outputs  # This is where syncing matters
    if stream is None:
        stream = session["stream"]
        path = session.outputs_path
        name = "Decoded Stream"
    else:
        path = session["array"].path
        name = session["array"].name

    if path is None:
        figure = html.p("Decoding Error: Invalid Byte-Stream Representation")
        array_size = ""
    else:
        figure = html.figure(
            html.img(src=URLOfPath(path)), html.figcaption(html.i(name))
        )
        array_size = path.stat().st_size

    output = html.div()
    with output:
        with html.table(style="margin-bottom:1em"):
            with html.tr():
                html.td(
                    html.div(
                        html.pre(stream),
                        style="margin-right:2em; height:10em; overflow:auto",
                    )
                )
                html.td(figure)
            with html.tr():
                html.td(f"Stream length: {stream_size}")
                html.td(f"Filesize: {array_size}")
        with html.table(border="1px", style="margin-bottom:1em"):
            with html.tr():
                for key in details.keys():
                    html.th(_LEGEND[key])
            with html.tr():
                for key, value in details.items():
                    if key in ("c", "v"):
                        html.td(html.div(value, style="width:10em; overflow:auto"))
                    elif key == "l":
                        html.td(
                            str(value)[1:-1].replace(",", " x"), _class="align_center"
                        )
                    else:
                        html.td(value, _class="align_center")

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Functions for the command line interface of PCA-B-Stream.
Run doctests below with: python cli.py

>>> import pathlib
>>> import sys
>>> from unittest.mock import patch
>>> path = pathlib.Path(".") / "resource" / "pca-0.png"
>>> with patch("sys.argv", new=["fake_cmd_name", path]):
...     PCA2BStream()
FnmHoFain+3jtU

>>> import imageio
>>> import numpy
>>> import pathlib
>>> import sys
>>> import tempfile
>>> from unittest.mock import patch
>>> folder = tempfile.mkdtemp()
>>> path = pathlib.Path(folder) / "a.png"
>>> with patch("sys.argv", new=["fake_cmd_name", "FnmHoFain+3jtU", path]):
...     BStream2PCA()
>>> original_path = pathlib.Path(".") / "resource" / "pca-0.png"
>>> original = imageio.imread(original_path)
>>> image = imageio.imread(path)
>>> print(numpy.array_equal(image, original))
True
"""

import sys as sstm
from pathlib import Path as path_t

import imageio as mgio

import pca_b_stream.main as pcas


def PCA2BStream() -> None:
    """"""
    error_code = -1

    if sstm.argv.__len__() != 2:
        print(
            f"{PCA2BStream.__name__.lower()}: No image specified or too many arguments"
        )
        sstm.exit(error_code)
    error_code -= 1

    path = path_t(sstm.argv[1])
    if not path.is_file():
        print(f"{path}: Specified path is not a(n existing) file")
        sstm.exit(error_code)
    error_code -= 1

    try:
        image = mgio.imread(path)
    except Exception as exception:
        print(exception)
        sstm.exit(error_code)
    error_code -= 1

    if not pcas.PCAIsValid(image):
        print(f"{path}: Not a valid Piecewise-Constant Array")
        sstm.exit(error_code)
    error_code -= 1

    print(pcas.PCA2BStream(image).decode("ascii"))


def BStream2PCA() -> None:
    """"""
    error_code = -1

    if sstm.argv.__len__() != 3:
        print(
            f"{BStream2PCA.__name__.lower()}: No stream and output file specified, or too many arguments"
        )
        sstm.exit(error_code)
    error_code -= 1

    stream = sstm.argv[1]
    if ("'" in stream) or ('"' in stream):
        print(
            f"{stream}: Stream contains \' or \"; "
            f"Note that the stream must not be passed with the \"b\" string type prefix"
        )
        sstm.exit(error_code)
    error_code -= 1

    stream = bytes(stream, "ascii")

    path = path_t(sstm.argv[2])
    if path.exists():
        print(
            f"{path}: Specified file already exists; Please delete first, or use another filename"
        )
        sstm.exit(error_code)
    error_code -= 1

    try:
        decoded = pcas.BStream2PCA(stream)
    except Exception as exception:
        print(exception)
        sstm.exit(error_code)
    error_code -= 1

    mgio.imwrite(path, decoded)


if __name__ == "__main__":
    #
    import doctest

    doctest.testmod()



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import dominate.tags as html

from pca_b_stream.flask.session.constants import PROJECT_NAME as NAME


NAME_MEANING = "Byte Stream Representation of Piecewise-constant Array"
SHORT_DESCRIPTION = (
    "Generation of a printable byte stream representation of a piecewise-constant Numpy array, "
    "and re-creation of the array from the byte stream."
)

ABOUT = html.div(
    html.h4("Purpose"),
    html.p(SHORT_DESCRIPTION),
    #
    html.h4("Privacy"),
    html.p(
        f"""
            Uploaded inputs are stored on the server for the sole processing by {NAME}.
            All data, uploaded, intermediate and final, can be deleted manually through a dedicated button
            that appears on processing completion.
            If not deleted manually, all data are automatically deleted based on their timestamp
            by a process running several times a day.
        """
    ),
    #
    html.h4("Links"),
    html.ul(
        html.li(
            html.a(
                "Documentation",
                href="https://gitlab.inria.fr/edebreuv/pca-b-stream/-/blob/master/README.rst",
            )
        ),
        html.li(
            html.a(
                "Source code (Gitlab repository)",
                href="https://gitlab.inria.fr/edebreuv/pca-b-stream/",
            )
        ),
        html.li(
            html.a(
                "Python Package Index (PyPI) page",
                href="https://pypi.org/project/pca-b-stream/",
            )
        ),
    ),
)

MAX_IMAGE_SIZE = 256  # MB



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from pathlib import Path as path_t
from typing import Any

import flask as flsk
import imageio as mgio
from si_fi_o.session.session import session_t

import pca_b_stream.main as pcas


output_types_h = tuple[tuple[str | None, int, Any], path_t | None]


def ProcessSession(
    session: session_t, session_id: str, folder: path_t, /
) -> output_types_h:
    """"""
    # FIXME: Find a better way to guess targeted conversion
    if "to byte stream" in flsk.request.form:
        return _ToSTR(session, session_id, folder)
    else:
        return _ToARY(session, session_id, folder)


def _ToSTR(session: session_t, _: str, __: path_t, /) -> output_types_h:
    """"""
    image = mgio.imread(session["array"].path)
    if pcas.PCAIsValid(image):
        stream = pcas.PCA2BStream(image)
        length = stream.__len__()
        details = pcas.BStreamDetails(
            stream,
            should_print=False,
            should_return=True,
        )
        stream = stream.decode("ascii")
    else:
        stream = "Invalid Piecewise-constant Array"
        length = -1
        details = None

    return (stream, length, details), None


def _ToARY(session: session_t, session_id: str, folder: path_t, /) -> output_types_h:
    """"""
    stream = bytes(session["stream"], "ascii")
    stream = b"\n".join(stream.splitlines())  # To convert newlines to \n

    try:
        decoded = pcas.BStream2PCA(stream)
    except:
        decoded = None

    length = stream.__len__()
    if decoded is None:
        details = None
        path = None
    else:
        details = pcas.BStreamDetails(
            stream,
            should_print=False,
            should_return=True,
        )
        path = folder / f"decoded-{session_id}.png"
        folder.mkdir(exist_ok=True)
        mgio.imwrite(path, decoded)

    return (None, length, details), path
